# Copyright 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from glanceclient.tests.unit.v2 import base
from glanceclient.tests import utils
from glanceclient.v2 import metadefs

NAMESPACE1 = 'Namespace1'
TAG1 = 'Tag1'
TAG2 = 'Tag2'
TAGNEW1 = 'TagNew1'
TAGNEW2 = 'TagNew2'
TAGNEW3 = 'TagNew3'


def _get_tag_fixture(tag_name, **kwargs):
    tag = {
        "name": tag_name
    }
    tag.update(kwargs)
    return tag


data_fixtures = {
    "/v2/metadefs/namespaces/%s/tags" % NAMESPACE1: {
        "GET": (
            {},
            {
                "tags": [
                    _get_tag_fixture(TAG1),
                    _get_tag_fixture(TAG2)
                ]
            }
        ),
        "POST": (
            {},
            {
                'tags': [
                    _get_tag_fixture(TAGNEW2),
                    _get_tag_fixture(TAGNEW3)
                ]
            }
        ),
        "DELETE": (
            {},
            {}
        )
    },
    "/v2/metadefs/namespaces/%s/tags/%s" % (NAMESPACE1, TAGNEW1): {
        "POST": (
            {},
            _get_tag_fixture(TAGNEW1)
        )
    },
    "/v2/metadefs/namespaces/%s/tags/%s" % (NAMESPACE1, TAG1): {
        "GET": (
            {},
            _get_tag_fixture(TAG1)
        ),
        "PUT": (
            {},
            _get_tag_fixture(TAG2)
        ),
        "DELETE": (
            {},
            {}
        )
    },
    "/v2/metadefs/namespaces/%s/tags/%s" % (NAMESPACE1, TAG2): {
        "GET": (
            {},
            _get_tag_fixture(TAG2)
        ),
    },
    "/v2/metadefs/namespaces/%s/tags/%s" % (NAMESPACE1, TAGNEW2): {
        "GET": (
            {},
            _get_tag_fixture(TAGNEW2)
        ),
    },
    "/v2/metadefs/namespaces/%s/tags/%s" % (NAMESPACE1, TAGNEW3): {
        "GET": (
            {},
            _get_tag_fixture(TAGNEW3)
        ),
    }

}

schema_fixtures = {
    "metadefs/tag": {
        "GET": (
            {},
            {
                "additionalProperties": True,
                "name": {
                    "type": "string"
                },
                "created_at": {
                    "type": "string",
                    "readOnly": True,
                    "description": ("Date and time of tag creation"),
                    "format": "date-time"
                },
                "updated_at": {
                    "type": "string",
                    "readOnly": True,
                    "description": ("Date and time of the last tag"
                                    "  modification"),
                    "format": "date-time"
                },
                'properties': {}
            }
        )
    }
}


class TestTagController(testtools.TestCase):
    def setUp(self):
        super(TestTagController, self).setUp()
        self.api = utils.FakeAPI(data_fixtures)
        self.schema_api = utils.FakeSchemaAPI(schema_fixtures)
        self.controller = base.BaseController(self.api, self.schema_api,
                                              metadefs.TagController)

    def test_list_tag(self):
        tags = self.controller.list(NAMESPACE1)
        actual = [tag.name for tag in tags]
        self.assertEqual([TAG1, TAG2], actual)

    def test_get_tag(self):
        tag = self.controller.get(NAMESPACE1, TAG1)
        self.assertEqual(TAG1, tag.name)

    def test_create_tag(self):
        tag = self.controller.create(NAMESPACE1, TAGNEW1)
        self.assertEqual(TAGNEW1, tag.name)

    def test_create_multiple_tags(self):
        properties = {
            'tags': [TAGNEW2, TAGNEW3]
        }
        tags = self.controller.create_multiple(NAMESPACE1, **properties)
        self.assertEqual([TAGNEW2, TAGNEW3], tags)

    def test_update_tag(self):
        properties = {
            'name': TAG2
        }
        tag = self.controller.update(NAMESPACE1, TAG1, **properties)
        self.assertEqual(TAG2, tag.name)

    def test_delete_tag(self):
        self.controller.delete(NAMESPACE1, TAG1)
        expect = [
            ('DELETE',
             '/v2/metadefs/namespaces/%s/tags/%s' % (NAMESPACE1, TAG1),
             {},
             None)]
        self.assertEqual(expect, self.api.calls)

    def test_delete_all_tags(self):
        self.controller.delete_all(NAMESPACE1)
        expect = [
            ('DELETE',
             '/v2/metadefs/namespaces/%s/tags' % NAMESPACE1,
             {},
             None)]
        self.assertEqual(expect, self.api.calls)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

import openstackdocstheme

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__),
                '..', '..')))

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'openstackdocstheme',
]

# openstackdocstheme options
repository_name = 'openstack/python-glanceclient'
bug_project = 'python-glanceclient'
bug_tag = ''

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'python-glanceclient'
copyright = u'OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
#html_theme = 'nature'
html_theme = 'openstackdocs'

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = ['_theme']
html_theme_path = [openstackdocstheme.get_html_theme_path()]

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

html_last_updated_fmt = '%Y-%m-%d %H:%M'

# -- Options for man page output ----------------------------------------------

# Grouping the document tree for man pages.
# List of tuples 'sourcefile', 'target', u'title', u'Authors name', 'manual'
man_pages = [
    ('cli/glance', 'glance', u'Client for OpenStack Images API',
     [u'OpenStack Foundation'], 1),
]



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
import testtools

from glanceclient import exc

HTML_MSG = """<html>
 <head>
   <title>404 Entity Not Found</title>
 </head>
 <body>
  <h1>404 Entity Not Found</h1>
  Entity could not be found
  <br /><br />
 </body>
</html>"""


class TestHTTPExceptions(testtools.TestCase):
    def test_from_response(self):
        """exc.from_response should return instance of an HTTP exception."""
        mock_resp = mock.Mock()
        mock_resp.status_code = 400
        out = exc.from_response(mock_resp)
        self.assertIsInstance(out, exc.HTTPBadRequest)

    def test_handles_json(self):
        """exc.from_response should not print JSON."""
        mock_resp = mock.Mock()
        mock_resp.status_code = 413
        mock_resp.json.return_value = {
            "overLimit": {
                "code": 413,
                "message": "OverLimit Retry...",
                "details": "Error Details...",
                "retryAt": "2014-12-03T13:33:06Z"
            }
        }
        mock_resp.headers = {
            "content-type": "application/json"
        }
        err = exc.from_response(mock_resp, "Non-empty body")
        self.assertIsInstance(err, exc.HTTPOverLimit)
        self.assertEqual("OverLimit Retry...", err.details)

    def test_handles_html(self):
        """exc.from_response should not print HTML."""
        mock_resp = mock.Mock()
        mock_resp.status_code = 404
        mock_resp.text = HTML_MSG
        mock_resp.headers = {
            "content-type": "text/html"
        }
        err = exc.from_response(mock_resp, HTML_MSG)
        self.assertIsInstance(err, exc.HTTPNotFound)
        self.assertEqual("404 Entity Not Found: Entity could not be found",
                         err.details)

    def test_format_no_content_type(self):
        mock_resp = mock.Mock()
        mock_resp.status_code = 400
        mock_resp.headers = {'content-type': 'application/octet-stream'}
        body = b'Error \n\n'
        err = exc.from_response(mock_resp, body)
        self.assertEqual('Error \n', err.details)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json
import six
import six.moves.urllib.parse as urlparse
import testtools

from glanceclient.v2 import schemas


class FakeAPI(object):
    def __init__(self, fixtures):
        self.fixtures = fixtures
        self.calls = []

    def _request(self, method, url, headers=None, data=None,
                 content_length=None):
        call = build_call_record(method, sort_url_by_query_keys(url),
                                 headers or {}, data)
        if content_length is not None:
            call = tuple(list(call) + [content_length])
        self.calls.append(call)

        fixture = self.fixtures[sort_url_by_query_keys(url)][method]

        data = fixture[1]
        if isinstance(fixture[1], six.string_types):
            try:
                data = json.loads(fixture[1])
            except ValueError:
                data = six.StringIO(fixture[1])

        return FakeResponse(fixture[0], fixture[1]), data

    def get(self, *args, **kwargs):
        return self._request('GET', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._request('POST', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._request('PUT', *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self._request('PATCH', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._request('DELETE', *args, **kwargs)

    def head(self, *args, **kwargs):
        return self._request('HEAD', *args, **kwargs)


class FakeSchemaAPI(FakeAPI):
    def get(self, *args, **kwargs):
        _, raw_schema = self._request('GET', *args, **kwargs)
        return schemas.Schema(raw_schema)


class RawRequest(object):
    def __init__(self, headers, body=None,
                 version=1.0, status=200, reason="Ok"):
        """A crafted request object used for testing.

        :param headers: dict representing HTTP response headers
        :param body: file-like object
        :param version: HTTP Version
        :param status: Response status code
        :param reason: Status code related message.
        """
        self.body = body
        self.status = status
        self.reason = reason
        self.version = version
        self.headers = headers

    def getheaders(self):
        return copy.deepcopy(self.headers).items()

    def getheader(self, key, default):
        return self.headers.get(key, default)

    def read(self, amt):
        return self.body.read(amt)


class FakeResponse(object):
    def __init__(self, headers=None, body=None,
                 version=1.0, status_code=200, reason="Ok"):
        """A crafted response object used for testing.

        :param headers: dict representing HTTP response headers
        :param body: file-like object
        :param version: HTTP Version
        :param status: Response status code
        :param reason: Status code related message.
        """
        self.body = body
        self.reason = reason
        self.version = version
        self.headers = headers
        self.headers['x-openstack-request-id'] = 'req-1234'
        self.status_code = status_code
        self.raw = RawRequest(headers, body=body, reason=reason,
                              version=version, status=status_code)

    @property
    def status(self):
        return self.status_code

    @property
    def ok(self):
        return (self.status_code < 400 or
                self.status_code >= 600)

    def read(self, amt):
        return self.body.read(amt)

    def close(self):
        pass

    @property
    def content(self):
        if hasattr(self.body, "read"):
            return self.body.read()
        return self.body

    @property
    def text(self):
        if isinstance(self.content, six.binary_type):
            return self.content.decode('utf-8')

        return self.content

    def json(self, **kwargs):
        return self.body and json.loads(self.text) or ""

    def iter_content(self, chunk_size=1, decode_unicode=False):
        while True:
            chunk = self.raw.read(chunk_size)
            if not chunk:
                break
            yield chunk

    def release_conn(self, **kwargs):
        pass


class TestCase(testtools.TestCase):
    TEST_REQUEST_BASE = {
        'config': {'danger_mode': False},
        'verify': True}


class FakeTTYStdout(six.StringIO):
    """A Fake stdout that try to emulate a TTY device as much as possible."""

    def isatty(self):
        return True

    def write(self, data):
        # When a CR (carriage return) is found reset file.
        if data.startswith('\r'):
            self.seek(0)
            data = data[1:]
        return six.StringIO.write(self, data)


class FakeNoTTYStdout(FakeTTYStdout):
    """A Fake stdout that is not a TTY device."""

    def isatty(self):
        return False


def sort_url_by_query_keys(url):
    """A helper function which sorts the keys of the query string of a url.

       For example, an input of '/v2/tasks?sort_key=id&sort_dir=asc&limit=10'
       returns '/v2/tasks?limit=10&sort_dir=asc&sort_key=id'. This is to
       prevent non-deterministic ordering of the query string causing
       problems with unit tests.
    :param url: url which will be ordered by query keys
    :returns url: url with ordered query keys
    """
    parsed = urlparse.urlparse(url)
    queries = urlparse.parse_qsl(parsed.query, True)
    sorted_query = sorted(queries, key=lambda x: x[0])

    encoded_sorted_query = urlparse.urlencode(sorted_query, True)

    url_parts = (parsed.scheme, parsed.netloc, parsed.path,
                 parsed.params, encoded_sorted_query,
                 parsed.fragment)

    return urlparse.urlunparse(url_parts)


def build_call_record(method, url, headers, data):
    """Key the request body be ordered if it's a dict type."""
    if isinstance(data, dict):
        data = sorted(data.items())
    if isinstance(data, six.string_types):
        # NOTE(flwang): For image update, the data will be a 'list' which
        # contains operation dict, such as: [{"op": "remove", "path": "/a"}]
        try:
            data = json.loads(data)
        except ValueError:
            return (method, url, headers or {}, data)
        data = [sorted(d.items()) for d in data]
    return (method, url, headers or {}, data)
